package cz.muni.pa165.component.service;

import cz.muni.pa165.common_library.dtos.CarComponentRequestDto;
import cz.muni.pa165.common_library.dtos.CarComponentResponseDto;
import cz.muni.pa165.common_library.dtos.CarResponseDto;
import cz.muni.pa165.common_library.exception.DatabaseException;
import cz.muni.pa165.common_library.knot.CommonDbGetter;
import cz.muni.pa165.component.data.model.CarComponent;
import cz.muni.pa165.component.data.repository.ComponentRepositoryInterface;
import cz.muni.pa165.component.mapper.CarComponentMapper;
import java.math.BigDecimal;
import java.util.List;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;


/**
 * Service for component initialization.
 */
@Service
public class ComponentService implements ComponentServiceInterface {

  private final ComponentRepositoryInterface componentRepository;
  private final CarComponentMapper carComponentMapper;
  private final CommonDbGetter dbGetter;

  @Autowired
  ComponentService(ComponentRepositoryInterface componentRepository,
                   CarComponentMapper carComponentMapper,
                   CommonDbGetter dbGetter) {
    this.componentRepository = componentRepository;
    this.carComponentMapper = carComponentMapper;
    this.dbGetter = dbGetter;
  }

  public CarComponentResponseDto postCarComponent(CarComponentRequestDto carComponentDto) {
    return carComponentMapper.carComponentConverter(componentRepository.save(
        carComponentMapper.carComponentDtoConverterWithoutId(carComponentDto)));
  }

  /**
   * Finds and returns car component by id.
   *
   * @param carComponentId id of the component.
   * @return found car component.
   */
  @Transactional(readOnly = true)
  public CarComponentResponseDto getCarComponentById(Long carComponentId) {
    return carComponentMapper.carComponentConverter(
        componentRepository.findById(carComponentId).orElseThrow(
          () -> new DatabaseException("Component with id" + carComponentId + "was not found.")));
  }

  /**
   * Get all the car components in the repository.
   *
   * @return list of stored car components.
   */
  @Transactional(readOnly = true)
  public List<CarComponentResponseDto> getAllCarComponents() {
    return componentRepository.findAll()
        .stream()
        .map(carComponentMapper::carComponentConverter)
        .toList();
  }

  /**
   * Removes car component from repository based on the id.
   *
   * @param carComponentId of the car component for removal.
   */
  @Transactional
  public String deleteById(Long carComponentId) {
    // search all cars if the component is not being used.
    var response = dbGetter.getAllCars();
    if (response.isPresent()) {
      List<CarResponseDto> cars = response.get();
      for (CarResponseDto car : cars) {
        if (car.getComponentIdsNames().contains(carComponentId)) {
          throw new DatabaseException("Component with id " + carComponentId
          + " cannot be removed as it is used by car with id " + car.getId());
        }
      }
    }
    // if no car has component assigned, remove it
    componentRepository.findById(carComponentId).orElseThrow(
        () -> new DatabaseException("Component with id " + carComponentId + " was not found."));
    componentRepository.deleteById(carComponentId);
    return "Component with id " + carComponentId + " was successfully deleted.";
  }

  /**
   * Saves a few car components.
   */
  @Transactional
  public void fillTheDataBase() {
    // name, weight, price, manufacturer
    // Tyres
    var heavyTyres = new CarComponent(0L,
        "Tyres",
        BigDecimal.valueOf(200),
        BigDecimal.valueOf(10000),
        "M");
    var expensiveTyres = new CarComponent(1L,
        "Tyres",
        BigDecimal.valueOf(150),
        BigDecimal.valueOf(23500),
        "T");
    var basicTyres = new CarComponent(2L,
        "Tyres",
        BigDecimal.valueOf(150),
        BigDecimal.valueOf(9000),
        "M");
    componentRepository.save(heavyTyres);
    componentRepository.save(expensiveTyres);
    componentRepository.save(basicTyres);

    // Engines
    var basicEngine = new CarComponent(3L,
        "Engine",
        BigDecimal.valueOf(350),
        BigDecimal.valueOf(150000),
        "H");
    var lightEngine = new CarComponent(4L,
        "Engine",
        BigDecimal.valueOf(320),
        BigDecimal.valueOf(230000),
        "H");
    var oldEngine = new CarComponent(5L,
        "Engine",
        BigDecimal.valueOf(350),
        BigDecimal.valueOf(100000),
        "M");
    componentRepository.save(basicEngine);
    componentRepository.save(lightEngine);
    componentRepository.save(oldEngine);

    // Gearboxes
    var basicGearbox = new CarComponent(6L,
        "Gearbox",
        BigDecimal.valueOf(70),
        BigDecimal.valueOf(12000),
        "GM");
    var extendedGearbox = new CarComponent(7L,
        "Gearbox",
        BigDecimal.valueOf(80),
        BigDecimal.valueOf(15000),
        "GM");
    var compactGearbox = new CarComponent(8L,
        "Gearbox",
        BigDecimal.valueOf(50),
        BigDecimal.valueOf(17000),
        "GM");
    componentRepository.save(basicGearbox);
    componentRepository.save(extendedGearbox);
    componentRepository.save(compactGearbox);
  }

  @Transactional
  public void clearDatabase() {
    componentRepository.deleteAll();
  }
}
